#include <unistd.h> /* fsync, close */
#include <stdio.h> /* rename */
#include "_stralloc.h"
#include "_buffer.h"
#include "_error.h"
#include "getln.h"
#include "_scan.h"
#include "bailout.h"
#include "_str.h"
#include "open.h"
#include "cdb_make.h"
#include "uogetopt.h"


static stralloc sa;
static void die_format(void) { bailout(0,"bad format: ",sa.s,0,0); }
static struct cdb_make cdbm;


static unsigned int
scanpair(const char *s, char sep, unsigned long *ul1, unsigned long *ul2,
	int *flagdone)
{
	unsigned int l;
	l=scan_ulong(s,ul1);
	if (l==0) die_format();
	if (s[l]=='-') {
		unsigned int m;
		l++;
		m=scan_ulong(s+l,ul2);
		if (m==0) die_format();
		l+=m;
	} else
		*ul2=*ul1;
	if (s[l]!=sep)
		die_format();
	else {
		l++;
		if (sep=='.' && s[l]==' ')
			*flagdone=1;
	}
	if (*ul2<*ul1) die_format();
	if (*ul2>255) die_format();
	if (*ul1>255) die_format();
	return l;
}

static void put1(char *key, unsigned int keylen,
	char *value, unsigned int vallen)
{
	if (-1==cdb_make_add(&cdbm, key,keylen,value,vallen))
		bailout(errno,"failed to add to cdb",0,0,0);
}

static void put(
	unsigned long u0a,  unsigned long u0b,
	unsigned long u1a,  unsigned long u1b,
	unsigned long u2a,  unsigned long u2b,
	unsigned long u3a,  unsigned long u3b,
	char *s)
{
	unsigned int sl=str_len(s);
	unsigned int i0,i1,i2,i3;
	char ip[4];
	if (u0a == 0 && u0b == 255 
			&& u1a == 0 && u1b == 255 
			&& u2a == 0 && u2b == 255 
			&& u3a == 0 && u3b==255) {
		/* /0 shortcut */
		put1(ip,0,s,sl);
		return;
	} 
	for (i0=u0a; i0 <= u0b; i0++)  {
		ip[0]=i0;

		if (u1a == 0 && u1b == 255 
				&& u2a == 0 && u2b == 255 
				&& u3a == 0 && u3b==255) {
			/* /8 shortcut */
			put1(ip,1,s,sl);
			continue;
		} 

		for (i1=u1a; i1 <= u1b; i1++)  {
			ip[1]=i1;

			if (u2a == 0 && u2b == 255 && u3a == 0 && u3b==255) {
				/* /16 shortcut */
				put1(ip,2,s,sl);
				continue;
			} 

			for (i2=u2a; i2 <= u2b; i2++) {
				ip[2]=i2;

				if (u3a == 0 && u3b==255) {
					/* /24 shortcut */
					put1(ip,3,s,sl);
					continue;
				} 
				for (i3=u3a; i3 <= u3b; i3++)  {
					ip[3]=i3;
					put1(ip,4,s,sl);
				}
			}
		}
	}
}

static void
read_classes(void)
{
    while (1) {
		int gotlf;
		unsigned int l;
		unsigned int i;
		unsigned long ul1a,ul1b;
		unsigned long ul2a,ul2b;
		unsigned long ul3a,ul3b;
		unsigned long ul4a,ul4b;
		int flagdone=0;
		if (-1==getln(buffer_0,&sa,&gotlf,'\n'))
		bailout(errno,"failed to read ","stdin",0,0);
		if (sa.len==0) break;
		if (!gotlf)
		bailout(0,"incomplete last line read from ","stdin",0,0);
		sa.s[--sa.len]=0;
		while (sa.len && (sa.s[sa.len-1]==' ' || sa.s[sa.len-1]=='\t'))
			sa.s[--sa.len]=0;
		i=scanpair(sa.s,'.',&ul1a,&ul1b,&flagdone);
		if (flagdone) {
			ul2a=0;
			ul2b=255;
		} else {
			l=scanpair(sa.s+i,'.',&ul2a,&ul2b,&flagdone);
			i+=l;
		}
		if (flagdone) {
			ul3a=0;
			ul3b=255;
		} else {
			l=scanpair(sa.s+i,'.',&ul3a,&ul3b,&flagdone);
			i+=l;
		}
		if (flagdone) {
			ul4a=0;
			ul4b=255;
		} else {
			l=scanpair(sa.s+i,' ',&ul4a,&ul4b,&flagdone);
			i+=l;
		}
		while (sa.s[i]==' ' || sa.s[i]=='\t') 
			i++;
		if (!sa.s[i]) die_format();
		put(ul1a,ul1b, ul2a,ul2b, ul3a,ul3b, ul4a,ul4b, sa.s+i);
	}
}

static uogetopt_t opts[]={
{0,"",UOGO_MINARGS | UOGO_HIDDEN,0, 3 , 0,0},
{0,"",UOGO_MAXARGS | UOGO_HIDDEN,0, 3 , 0,0},
{'\0',0,UOGO_FLAG,0,0,0,0}
};

int
main(int argc, char **argv)
{
	int fd;

	bailout_progname(argv[0]);
	uogetopt(flag_bailout_log_name,PACKAGE,VERSION,
		&argc,argv,uogetopt_out,
		"usage: compile-classes CDB TMP <input", opts,0);

	fd=open_trunc(argv[2]);
	if (fd==-1) bailout(errno,"failed to open ",argv[2],0,0);
	if (-1==cdb_make_start (&cdbm,fd))
		bailout(errno,"failed to write ",argv[2],0,0);
	read_classes();
	if (-1==cdb_make_finish(&cdbm))
		bailout(errno,"cdb_make_finish failed",0,0,0);
	if (fsync (fd) == -1) bailout (errno, "failed to fsync",argv[2],0,0);
	if (close (fd) == -1) bailout (errno, "failed to fclose",argv[2],0,0);
	if (rename (argv[2],argv[1]))
		bailout(errno,"failed to rename ",argv[2]," to ",argv[1]);

	_exit(0);
}
